/*****************************************************************************
SQL Server 2016: How to have local temporary tables, global temporary tables, 
table variables and table-valued parameters use In-Memory technology instead 
of being bound to tempdb.
Source: https://blogs.msdn.microsoft.com/sqlserverstorageengine/2016/03/21/improving-temp-table-and-table-variable-performance-using-memory-optimization/

Testing uses ostress, part of the RM Utilities for SQL Server. Download from:
x64: https://www.microsoft.com/en-us/download/details.aspx?id=4511
x32: https://www.microsoft.com/en-us/download/details.aspx?id=8161
*****************************************************************************/

SET NOCOUNT ON;
GO
USE master;
GO
RAISERROR('Dropping / creating InMemoryTempTables database', 10, 1) WITH NOWAIT;
GO
DROP DATABASE IF EXISTS InMemoryTempTables;
GO
CREATE DATABASE InMemoryTempTables;
GO
USE InMemoryTempTables;
GO
/*****************************************************************************
Next we need to add InMemory OLTP support to this database. This is done by
adding a new filegroup that is specified to hold memory optimized data.
Then, a file (actually, a path) will be added to this filegroup. This path is 
at the location of where the instance is specified to have the default data.
*****************************************************************************/
ALTER DATABASE CURRENT ADD FILEGROUP [mod] CONTAINS MEMORY_OPTIMIZED_DATA;
-- Because too many SQL commands don't take variables, we need to build a command into a string and then execute it.
--SELECT SERVERPROPERTY('InstanceDefaultDataPath');
DECLARE @sql nvarchar(max) = N'ALTER DATABASE CURRENT
       ADD FILE (name=''mod'', filename=' +
       QUOTENAME(cast(SERVERPROPERTY('InstanceDefaultDataPath') as nvarchar(max)) + db_name() + N'_mod', N'''') +
             ') TO FILEGROUP [mod]';
EXECUTE sp_executesql @sql;
GO

/*****************************************************************************
First, let's memory optimize a local temp table.
Create a stored procedure using a regular #temp table.
*****************************************************************************/
RAISERROR('Dropping / creating dbo.sp_temp', 10, 1) WITH NOWAIT;
GO
-- the normal temp table way:
DROP PROCEDURE IF EXISTS dbo.sp_temp;
GO
CREATE PROCEDURE dbo.sp_temp
AS
BEGIN
    DROP TABLE IF EXISTS #temp1;
    CREATE TABLE #temp1
        (
            c1 INT NOT NULL ,
            c2 NVARCHAR(4000)
        );
    BEGIN TRAN
    DECLARE @i INT = 0;
    WHILE @i < 100
        BEGIN
            INSERT  #temp1
            VALUES  ( @i, N'abc' );
            SET @i += 1
        END;
    COMMIT;
END;
GO

/******************************************************************************
Create a second procedure that will use a memory optimized table. In order to 
use the memory optimized table across multiple sessions, where each sees just 
their data, we need to add a Security Policy (row-level security) to restrict 
the session to just the rows in the table for their session. The table needs to
have a session_id column (default @@SPID), with a check constraint (more 
performant that a blocking security predicate). The security policy uses a
function that returns whether the passed in session_id is for the current @@SPID.
******************************************************************************/
RAISERROR('Dropping / creating dbo.fn_SessionFilter', 10, 1) WITH NOWAIT;
GO
-- This single filter function can be used for all session-level "temp" tables.
DROP FUNCTION IF EXISTS dbo.fn_SessionFilter;
GO
CREATE FUNCTION dbo.fn_SessionFilter(@session_id smallint)
    RETURNS TABLE
WITH SCHEMABINDING, NATIVE_COMPILATION
AS
    RETURN SELECT 1 as fn_SessionFilter WHERE @session_id=@@SPID;
GO

-- Create a memory optimized schema-only table that will be used to simulate a #temp table.
-- There needs to be a differently named table for each different schema type (no more having all procs using #temp).
RAISERROR('Dropping / creating dbo.temp1', 10, 1) WITH NOWAIT;
GO
DROP TABLE IF EXISTS dbo.temp1;
GO
CREATE TABLE dbo.temp1
    (
      c1 INT NOT NULL ,
      c2 NVARCHAR(4000) ,
      session_id SMALLINT NOT NULL DEFAULT ( @@SPID ) ,
      INDEX IX_session_id ( session_id ) ,
      CONSTRAINT CHK_temp1_session_id CHECK ( session_id = @@SPID ),
    )
    WITH (MEMORY_OPTIMIZED=ON, DURABILITY=SCHEMA_ONLY);
GO

RAISERROR('Dropping / creating dbo.temp1Filter', 10, 1) WITH NOWAIT;
GO
-- Add the security policy to enable filtering on session_id. 
-- A different security policy is needed for each temp table.
DROP SECURITY POLICY IF EXISTS dbo.temp1Filter;
GO
CREATE SECURITY POLICY dbo.temp1Filter
ADD FILTER PREDICATE dbo.fn_SessionFilter(session_id)
ON dbo.temp1
WITH (STATE = ON);
GO

RAISERROR('Dropping / creating dbo.sp_temp_mo', 10, 1) WITH NOWAIT;
GO
-- Build a stored procedure that utilizes a memory optimized "temp" table.
DROP PROCEDURE IF EXISTS dbo.sp_temp_mo;
GO
CREATE PROCEDURE dbo.sp_temp_mo
AS
BEGIN
	-- The security policy will restrict this session to only rows available to this session
	-- This delete will delete any possible existing rows for this session_id (spid)
    DELETE FROM dbo.temp1;
    BEGIN TRAN
    DECLARE @i INT = 0;
    WHILE @i < 100
        BEGIN
            INSERT  dbo.temp1 (c1, c2)
            VALUES  ( @i, N'abc' );
            SET @i += 1;
        END;
    COMMIT;
END;
GO

/******************************************************************************
Testing:
1. If using a named instance, replace "-S." with "-S.\<instance name>"
Test regular temp tables with:
ostress -S. -dInMemoryTempTables -q -n100 -r1000 -Q"sp_temp"
: 100 concurrent connections, 1,000 executions (total 100,000 executions)
my results: 00:01:59.526

Then test memory-optimzed "temp" tables with:
ostress -S. -dInMemoryTempTables -q -n100 -r1000 -Q"sp_temp_mo"
my results: 00:00:40.651
******************************************************************************/





/******************************************************************************
Memory optimizing global temp tables:
First, a stored procedure using a regular global temp table.
******************************************************************************/
RAISERROR('Dropping / creating dbo.sp_gtemp', 10, 1) WITH NOWAIT;
GO
DROP PROCEDURE IF EXISTS dbo.sp_gtemp;
GO
CREATE PROCEDURE dbo.sp_gtemp
AS
BEGIN
    IF NOT EXISTS (SELECT * FROM tempdb.sys.objects WHERE name=N'##temp1')
            CREATE TABLE ##temp1
                (
                    c1 INT NOT NULL ,
                    c2 NVARCHAR(4000)
                );
    BEGIN TRAN;
    DECLARE @i INT = 0;
    WHILE @i < 100
        BEGIN
            INSERT  ##temp1
            VALUES  ( @i, N'abc' );
            SET @i += 1;
        END;
    COMMIT;
END;
GO

/******************************************************************************
Memory optimizing global temp tables:
Now, build a memory optimized table, and a stored procedure to utilize it.
******************************************************************************/
RAISERROR('Dropping / creating dbo.gtemp1', 10, 1) WITH NOWAIT;
GO
DROP TABLE IF EXISTS dbo.gtemp1;
GO
CREATE TABLE dbo.gtemp1
(c1 INT NOT NULL INDEX ix_1 ,
 c2 NVARCHAR(4000))
WITH (MEMORY_OPTIMIZED=ON, DURABILITY=SCHEMA_ONLY);
GO

RAISERROR('Dropping / creating dbo.sp_gtemp_mo', 10, 1) WITH NOWAIT;
GO
DROP PROCEDURE IF EXISTS dbo.sp_gtemp_mo
GO
CREATE PROCEDURE sp_gtemp_mo
AS
    BEGIN
        BEGIN TRAN;
        DECLARE @i INT = 0;
        WHILE @i < 100
            BEGIN
                INSERT  dbo.gtemp1
                VALUES  ( @i, N'abc' );
                SET @i += 1;
            END;
        COMMIT;
    END;
GO

/******************************************************************************
Testing:
1. If using a named instance, replace "-S." with "-S.\<instance name>"
Test global temp tables with:
ostress -S. -dInMemoryTempTables -q -n100 -r1000 -Q"sp_gtemp"
: 100 concurrent connections, 1,000 executions (total 100,000 executions)
my results: 00:01:43.142

Then test memory-optimzed "temp" tables with:
ostress -S. -dInMemoryTempTables -q -n100 -r1000 -Q"sp_gtemp_mo"
my results: 00:00:34.951
******************************************************************************/





/******************************************************************************
In order to have memory optimized table variables or table parameters, they 
both need to have a memory optimized table type. Memory optimized tables must
have an index, and the table declaration needs to specify memory optimized.
******************************************************************************/
RAISERROR('Dropping / creating dbo.test_disk', 10, 1) WITH NOWAIT;
GO
-- disk based table type:
DROP TYPE IF EXISTS dbo.test_disk;
CREATE TYPE dbo.test_disk AS TABLE
(c1 INT NOT NULL,
 c2 CHAR(10));

 RAISERROR('Dropping / creating dbo.test_memory', 10, 1) WITH NOWAIT;
GO
-- memory optimized table type
DROP TYPE IF EXISTS dbo.test_memory;
CREATE TYPE dbo.test_memory AS TABLE
(c1 INT NOT NULL INDEX ix_c1,
 c2 CHAR(10))
WITH (MEMORY_OPTIMIZED=ON);
GO


RAISERROR('Dropping / creating dbo.sp_tvar', 10, 1) WITH NOWAIT;
GO
DROP PROCEDURE IF EXISTS dbo.sp_tvar;
GO
CREATE PROCEDURE dbo.sp_tvar 
AS
DECLARE @tv dbo.test_disk;
INSERT @tv VALUES (1, 'n');
INSERT @tv VALUES (2, 'm');
DELETE FROM @tv;
GO


RAISERROR('Dropping / creating dbo.sp_tvar_mo', 10, 1) WITH NOWAIT;
GO
DROP PROCEDURE IF EXISTS dbo.sp_tvar_mo;
GO
CREATE PROCEDURE dbo.sp_tvar_mo
AS
DECLARE @tv dbo.test_memory;
INSERT @tv VALUES (1, 'n');
INSERT @tv VALUES (2, 'm');
DELETE FROM @tv;
GO


/******************************************************************************
TESTING: Run the following code

SET NOCOUNT ON;
DROP TABLE IF EXISTS #times;
CREATE TABLE #times (StartTime DATETIME2 DEFAULT (SYSDATETIME()), EndTime DATETIME2);
INSERT INTO #times DEFAULT VALUES;
GO
EXECUTE sp_tvar;
GO 10000
UPDATE #times SET EndTime = SYSDATETIME() WHERE EndTime IS NULL;
/*
My results:
# CPUS	# Loops		
1		10,000		5,656
1		50,000		30,734
4		10,000		7,250
4		50,000		35,562
*/

INSERT INTO #times DEFAULT VALUES;
GO
EXECUTE sp_tvar_mo;
GO 10000
UPDATE #times SET EndTime = SYSDATETIME() WHERE EndTime IS NULL;
SELECT *, DATEDIFF(MILLISECOND, StartTime, EndTime) AS [Elapsed (ms)] FROM #times ORDER BY StartTime;
/*
My results:
# CPUS	# Loops		
1		10,000		2,547
1		50,000		12,626
4		10,000		3,937
4		50,000		19,516
*/

Now, test with ostress:
Testing:
1. If using a named instance, replace "-S." with "-S.\<instance name>"
Test global temp tables with:
ostress -S. -dInMemoryTempTables -q -n100 -r1000 -Q"sp_tvar"
: 100 concurrent connections, 1,000 executions (total 100,000 executions)
my results: 00:00:49.050

Then test memory-optimzed "temp" tables with:
ostress -S. -dInMemoryTempTables -q -n100 -r1000 -Q"sp_trav_mo"
my results: 00:00:09.303

******************************************************************************/
